#include "aimtrans.h"

void *at_session_main(void *arg);

static int getaimdata(char *aimbinarydir, unsigned char *buf, int buflen, unsigned long offset, char *modname)
{
  FILE *f;
  static const char defaultmod[] = "aim.exe";
  char *filename = NULL;

  if (modname) {

    if (!(filename = malloc(strlen(aimbinarydir)+1+strlen(modname)+4+1))) {
      return -1;
    }

    sprintf(filename, "%s/%s.ocm", aimbinarydir, modname);

  } else {

    if (!(filename = malloc(strlen(aimbinarydir)+1+strlen(defaultmod)+1))) {
      return -1;
    }

    sprintf(filename, "%s/%s", aimbinarydir, defaultmod);

  }

  printf("\n\nREADING %d at %d from %s\n\n", buflen, offset, filename);
  if (!(f = fopen(filename, "r"))) {
    free(filename);
    return -1;
  }

  free(filename);

  if (fseek(f, offset, SEEK_SET) == -1) {
    fclose(f);
    return -1;
  }

  if (fread(buf, buflen, 1, f) != 1) {
    fclose(f);
    return -1;
  }

  fclose(f);

  return buflen;
}

static int at_memrequest(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  va_list ap;
  unsigned long offset, len;
  unsigned char *buf;
  char *modname;
  int j;
  at_session s;
  ati ti;

  s = (at_session)sess->aux_data;
  ti = s->ti;
  
  va_start(ap, command);
  offset = va_arg(ap, unsigned long);
  len = va_arg(ap, unsigned long);
  modname = va_arg(ap, char *);
  va_end(ap);


  log_debug(ZONE, "We got a memrequest\n");
  if(ti->offset == offset && ti->read_len == len && 
                  strcmp(ti->modname, modname) == 0 && ti->send_buf)
  {
    printf("\n\nSENDING CACHE DATA\n\n");
    aim_sendmemblock(sess, command->conn, offset, len, ti->send_buf);
  } else {
    if (!(buf = malloc(len))) {
        return 0;
    }
    
    printf("\n\nREADING %d at %d from %s\n\n", len, offset, modname);
    if (getaimdata(ti->aimbinarydir, buf, len, offset, modname) == len) {
      ti->offset = offset;
      ti->read_len = len;
      if(ti->modname != NULL)
         free(ti->modname);
      ti->modname = strdup(modname);
      if(ti->send_buf != NULL)
         free(ti->send_buf);
      ti->send_buf = strdup(buf);
      aim_sendmemblock(sess, command->conn, offset, len, buf);
    } else {
      ti->offset = offset;
      ti->read_len = len;
      if(ti->modname != NULL)
         free(ti->modname);
      ti->modname = NULL;
      if(ti->send_buf != NULL)
         free(ti->send_buf);
      ti->send_buf = NULL;
      aim_sendmemblock(sess, command->conn, offset, len, NULL);
    }
    
    free(buf);
  }

  return 1;
}

int at_rateresp(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
    char *buddylist;
    at_session s;

    s = sess->aux_data;


    switch(command->conn->type) {
        case AIM_CONN_TYPE_BOS: {
            /* this is the new buddy list */
            char buddies[] = "";
            /* this is the new profile */
            /* XXX Can we let people do this? */
            char profile[] = "";  

            aim_bos_ackrateresp(sess, command->conn); 
            aim_bos_reqpersonalinfo(sess, command->conn);
            aim_bos_reqlocaterights(sess, command->conn);
            /* XXX evaluate these again later */
            aim_bos_setprofile(sess, command->conn, profile, NULL, AIM_CAPS_CHAT | AIM_CAPS_BUDDYICON);
            aim_bos_reqbuddyrights(sess, command->conn);

            /* send the buddy list and profile (required, even if empty) */
            buddylist = at_buddy_buildlist(s, s->cur);
            if(buddylist==NULL)
            {
                log_debug(ZONE, "[AIM] We have a NULL buddylist");
                if(aim_bos_setbuddylist(sess, command->conn, buddies)<0)
                    log_debug(ZONE, "[AIM] Error setting buddylist!");
            } else {
                log_debug(ZONE, "[AIM] Setting buddylist: %s", buddylist);
                if(aim_bos_setbuddylist(sess, command->conn, buddylist)<0)
                    log_debug(ZONE, "[AIM] Error setting buddylist!");
                free(buddylist);
            }

            /* dont really know what this does */
            aim_addicbmparam(sess, command->conn);
            aim_bos_reqicbmparaminfo(sess, command->conn);  
  
            aim_bos_reqrights(sess, command->conn);  
            /* set group permissions -- all user classes */
            /* XXX evaluate these! */
            aim_bos_setgroupperm(sess, command->conn, AIM_FLAG_ALLUSERS);
            aim_bos_setprivacyflags(sess, command->conn, 
                                AIM_PRIVFLAGS_ALLOWIDLE);

            break;  
        }
        case AIM_CONN_TYPE_AUTH:
            aim_bos_ackrateresp(sess, command->conn);
            aim_auth_clientready(sess, command->conn);
            log_debug(ZONE, "[AT] connected to authorization/admin service");
            break;

        default: 
            log_debug(ZONE, "[AIM] got rate response for unhandled connection type %04x", command->conn->type);
            break;
    }

    return 1;
}

int at_bosrights(struct aim_session_t *sess, 
                 struct command_rx_struct *command, ...)
{
    unsigned short maxpermits, maxdenies;
    va_list ap;
    xmlnode x;
    at_session s;
    ati ti;

    s = sess->aux_data;
    ti = s->ti;

    va_start(ap, command);
    maxpermits = va_arg(ap, u_int);
    maxdenies = va_arg(ap, u_int);
    va_end(ap);

    aim_bos_clientready(sess, command->conn);

    log_debug(ZONE, "[AIM] officially connected to BOS, sending pres.");

    s->loggedin = 1;
    x = jutil_presnew(JPACKET__AVAILABLE, jid_full(s->cur), "Connected");
    xmlnode_put_attrib(x, "from", jid_full(s->from));
    deliver(dpacket_new(x), ti->i);

    return 1;
}

int at_parse_login(struct aim_session_t *sess, 
                   struct command_rx_struct *command, ...)
{
    /* GONE - 10-21-01
     struct client_info_s info = {
           "AOL Instant Messenger (SM), version 4.1.2010/WIN32", 
            4, 1, 2010, "us", "en", 0x0004, 0x0000,  0x0000004b}; 
    */
    /* GONE - 10-22-01 
     * FIXED - with libfaim hack
     */
     //struct client_info_s info = {"AOL Instant Messenger (SM), version 4.1.2010/WIN32", 4, 30, 3141, "us", "en", 0x0004, 0x0001, 0x00000055};
    /* Actual WinAIM values 4.3.229 version */
    /* Doesn't Have Buddy Pres
     * GONE - 10-22-01
     * FIXED - with libfiam hack
     */
     //struct client_info_s info = {"AOL Instant Messenger (SM), version 4.3.2229/WIN32", 4, 3, 2229, "us", "en", 0x0109, 0x0000, 0x00000008c};
     // Works with pres
     /*struct client_info_s info = {"AOL Instant Messenger (SM), version 4.3.2229/WIN32", 4, 3, 2229, "us", "en", 0x0004, 0x0000, 0x0000008c};*/
     // Still Works, moving to libfaim stuff
     // struct client_info_s info = {"AOL Instant Messenger (SM), version 4.2.2110/WIN32",  4, /* major -- 0x0017 */ 2, /* minor -- 0x0018 */ 0x083e, /* build -- 0x001a */ "us", "en", 0x0109, /* major2 -- 0x0016 */ 0x0000, /* minor2 -- 0x0019 */ 0x0000007c /* unknown -- 0x0014 */}; /* 4.2.2110 */
    struct client_info_s info = AIM_CLIENTINFO_KNOWNGOOD;
    
    unsigned char *key;
    va_list ap;
    at_session s;
  
    s = (at_session)sess->aux_data;
    va_start(ap, command);
    key = va_arg(ap, char *);
    va_end(ap);

    aim_send_login(sess,command->conn,s->screenname,s->password, &info, key);
 
    return 1;
}

int at_parse_authresp(struct aim_session_t *sess, 
                      struct command_rx_struct *command, ...)
{
    va_list ap;
    struct aim_conn_t *bosconn = NULL;
    at_mio am;
    char *sn = NULL, *bosip = NULL, *errurl = NULL, *email = NULL;
    unsigned char *cookie = NULL;
    int errorcode = 0, regstatus = 0;
    int latestbuild = 0, latestbetabuild = 0;
    char *latestrelease = NULL, *latestbeta = NULL;
    char *latestreleaseurl = NULL, *latestbetaurl = NULL;
    char *latestreleaseinfo = NULL, *latestbetainfo = NULL;
    at_session s;
    struct aim_conn_t *conn = NULL;
    ati ti;

    s = (at_session)sess->aux_data;
    ti = s->ti;

    va_start(ap, command);
    sn = va_arg(ap, char *);
    errorcode = va_arg(ap, int);
    errurl = va_arg(ap, char *);
    regstatus = va_arg(ap, int);
    email = va_arg(ap, char *);
    bosip = va_arg(ap, char *);
    cookie = va_arg(ap, unsigned char *);

    latestrelease = va_arg(ap, char *);
    latestbuild = va_arg(ap, int);
    latestreleaseurl = va_arg(ap, char *);
    latestreleaseinfo = va_arg(ap, char *);

    latestbeta = va_arg(ap, char *);
    latestbetabuild = va_arg(ap, int);
    latestbetaurl = va_arg(ap, char *);
    latestbetainfo = va_arg(ap, char *);

    conn = command->conn;

    va_end(ap);

    log_debug(ZONE, "Auth Response for Screen name: %s\n", sn);

    /*
    * Check for error.
    */
    if (errorcode || !bosip || !cookie)
    {
        spool sp;
        xmlnode x;
        xmlnode body;
        char *errorstr;

        x = xmlnode_new_tag("message");
        xmlnode_put_attrib(x, "to", jid_full(s->cur));
        xmlnode_put_attrib(x, "from", jid_full(s->from));
        xmlnode_put_attrib(x, "type", "error");
        body = xmlnode_insert_tag(x, "body");

        sp = spool_new(xmlnode_pool(x));
        spooler(sp, "Error Code #%04x While Logging in to AIM\n",
                "Error URL: %s\0", sp);
        errorstr = pmalloc(xmlnode_pool(x), 200);
        snprintf(errorstr, 200, spool_print(sp), errorcode, errurl);
        xmlnode_insert_cdata(body, errorstr, strlen(errorstr));
        deliver(dpacket_new(x), ti->i);

        s->loggedin = 0;
        /* XXX End the Stuff */
        aim_conn_kill(sess, &command->conn);
        return 1;
    }

    aim_conn_kill(sess, &command->conn);
    bosconn = aim_newconn(sess, AIM_CONN_TYPE_BOS, bosip);
    /* XXX Handle these better so we actually return and free the threads */
    if (bosconn == NULL) {
        fprintf(stderr, "at: could not connect to BOS: internal error\n");
    } else if (bosconn->status & AIM_CONN_STATUS_CONNERR) {	
        fprintf(stderr, "at: could not connect to BOS\n");
        aim_conn_kill(sess, &bosconn);
    } else {

        aim_conn_setlatency(bosconn, 1);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, 
                            AIM_CB_SPECIAL_CONNCOMPLETE, at_conncomplete, 0);
        aim_conn_addhandler(sess, bosconn, 0x0009, 0x0003, 
                            at_bosrights, 0);
        aim_conn_addhandler(sess, bosconn, 0x0001, 0x0007, 
                            at_rateresp, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_ACK, 
                            AIM_CB_ACK_ACK, NULL, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_SERVERREADY, at_serverready, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_RATEINFO, at_rateresp, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_REDIRECT, at_handleredirect, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, 
                            AIM_CB_BUD_ONCOMING, at_parse_oncoming, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, 
                            AIM_CB_BUD_OFFGOING, at_parse_offgoing, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, 
                            AIM_CB_MSG_INCOMING, at_parse_incoming_im, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, 
                            AIM_CB_LOC_ERROR, at_parse_locerr, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, 
                            AIM_CB_MSG_MISSEDCALL, at_parse_misses, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_RATECHANGE, at_parse_ratechange, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_EVIL, at_parse_evilnotify, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, 
                            AIM_CB_MSG_ERROR, at_parse_msgerr, 0);
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, 
                            AIM_CB_LOC_USERINFO, at_parse_userinfo, 0);
        /* XXX Future?
         * aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, 
                            AIM_CB_MSG_ACK, at_parse_msgack, 0);
         */

        /*
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_CTN, 
                            AIM_CB_CTN_DEFAULT, aim_parse_unknown, 0);
        */
        /*
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, 
                            AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
        */
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, 
                            AIM_CB_GEN_MOTD, at_parse_motd, 0);
        
        aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, 
                            AIM_CB_SPECIAL_CONNERR, at_parse_connerr, 0);
        aim_conn_addhandler(sess, bosconn, 0x0001, 0x001f, 
                            at_memrequest, 0);
        
    
        am = pmalloco(s->p, sizeof(_at_mio));
        am->p = s->p;
        am->s = s;
        am->ass = sess;
        am->conn = bosconn;

        //m = mio_new(bosconn->fd, &at_handle_mio, am, mio_handlers_new(at_session_handle_read, at_session_handle_write, NULL));

        //s->m = m;

        aim_auth_sendcookie(sess, bosconn, cookie);

    }
    
    return 1;
}

int at_serverready(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  switch (command->conn->type)
  {
    case AIM_CONN_TYPE_AUTH:
        aim_auth_setversions(sess, command->conn);
        aim_bos_reqrate(sess, command->conn); /* request rate info */

        log_debug(ZONE, "[AT] done with auth ServerReady");
        break;

    case AIM_CONN_TYPE_BOS:

      aim_setversions(sess, command->conn);
      aim_bos_reqrate(sess, command->conn); /* request rate info */

      break;

      /* XXX Future expansion */
/*    case AIM_CONN_TYPE_CHATNAV:
      fprintf(stderr, "faimtest: chatnav: got server ready\n");
      aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CTN, AIM_CB_CTN_INFO, faimtest_chatnav_info, 0);
      aim_bos_reqrate(sess, command->conn);
      aim_bos_ackrateresp(sess, command->conn);
      aim_chatnav_clientready(sess, command->conn);
      aim_chatnav_reqrights(sess, command->conn);

      break;
    case AIM_CONN_TYPE_CHAT:
      aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERJOIN, faimtest_chat_join, 0);
      aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERLEAVE, faimtest_chat_leave, 0);
      aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_ROOMINFOUPDATE, faimtest_chat_infoupdate, 0);
      aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_INCOMINGMSG, faimtest_chat_incomingmsg, 0);
      aim_bos_reqrate(sess, command->conn);
      aim_bos_ackrateresp(sess, command->conn);
      aim_chat_clientready(sess, command->conn);
      break;

    case AIM_CONN_TYPE_RENDEZVOUS: 
      break;
*/
    default:
      log_debug(ZONE, "[AT] unknown connection type on Server Ready\n");
    }
  return 1;
}

int at_handleredirect(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  va_list ap;
  int serviceid;
  char *ip;
  unsigned char *cookie;
  at_session s;
  at_mio am;

  va_start(ap, command);
  serviceid = va_arg(ap, int);
  ip = va_arg(ap, char *);
  cookie = va_arg(ap, unsigned char *);

  s = (at_session)sess->aux_data;
 
  switch(serviceid)
  {
    case 0x0005:
    {
    	struct aim_conn_t *tstconn;
            
        tstconn = aim_newconn(sess, AIM_CONN_TYPE_ADS, ip);
	    if ((tstconn==NULL) || (tstconn->status & AIM_CONN_STATUS_RESOLVERR)) {
        	  log_debug(ZONE, "[AT] unable to reconnect with authorizer");
      	} else {
	      aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_FLAPVER, at_flapversion, 0);
    	  aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNCOMPLETE, at_conncomplete, 0);
	      aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, at_serverready, 0);
	      aim_conn_addhandler(sess, tstconn, 0x0001, 0x0007, at_rateresp, 0); 
    	  //aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_GEN, 0x0018, faimtest_hostversions, 0);
	      aim_auth_sendcookie(sess, tstconn, cookie);
    	  log_debug(ZONE, "sent cookie to adverts host");
       }
	}
    break;
    case 0x0007: /* Authorizer */
      {
	struct aim_conn_t *tstconn;
	/* Open a connection to the Auth */
	tstconn = aim_newconn(sess, AIM_CONN_TYPE_AUTH, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR) )
    {
	    log_debug(ZONE, "[AT] unable to reconnect with authorizer\n");
    }
	else
    {
        aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_FLAPVER, at_flapversion, 0);
        aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNCOMPLETE, at_conncomplete, 0);
        aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, at_serverready, 0);
	    aim_conn_addhandler(sess, tstconn, 0x0001, 0x0007, at_rateresp, 0);
	  /* Send the cookie to the Auth */
/*
        am = pmalloco(s->p, sizeof(_at_mio));
        am->p = s->p;
        am->s = s;
        am->conn = tstconn;
        am->ass = sess;
*/
        
        //mio_new(tstconn->fd, &at_handle_mio, tstconn, mio_handlers_new(at_session_handle_read, at_session_handle_write, NULL));

	    aim_auth_sendcookie(sess, tstconn, cookie);
    }

      }  
      break;
    case 0x000d: /* XXX ChatNav */
      {
	/*struct aim_conn_t *tstconn = NULL;
	tstconn = aim_newconn(sess, AIM_CONN_TYPE_CHATNAV, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR)) {
	  fprintf(stderr, "faimtest: unable to connect to chatnav server\n");
	  if (tstconn) aim_conn_kill(sess, &tstconn);
	  return 1;
	}
	aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, faimtest_serverready, 0);
	aim_auth_sendcookie(sess, tstconn, cookie);
	fprintf(stderr, "\achatnav: connected\n");*/
      }
      break;
    case 0x000e: /* Chat */
      {
	/*char *roomname = NULL;
	struct aim_conn_t *tstconn = NULL;

	roomname = va_arg(ap, char *);

	tstconn = aim_newconn(sess, AIM_CONN_TYPE_CHAT, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
	  {
	    fprintf(stderr, "faimtest: unable to connect to chat server\n");
	    if (tstconn) aim_conn_kill(sess, &tstconn);
	    return 1;
	  }		
	printf("faimtest: chat: connected\n");

	aim_chat_attachname(tstconn, roomname);

	aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, faimtest_serverready, 0);
	aim_auth_sendcookie(sess, tstconn, cookie);*/
      }
      break;
    default:
      log_debug(ZONE, "uh oh... got redirect for unknown service 0x%04x!!\n", serviceid);
      /* dunno */
    }

  va_end(ap);

  return 1;
}

int at_parse_ratechange(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
/* XXX Do we want to deal with this? */
/*  va_list ap;
  unsigned long newrate;
  
  va_start(ap, command); 
  newrate = va_arg(ap, unsigned long);
  va_end(ap);

  printf("faimtest: ratechange: %lu\n", newrate);
*/
  return 1;
}

int at_parse_motd(struct aim_session_t *sess, 
                  struct command_rx_struct *command, ...)
{
/* XXX Should we do the headline?
  static char *codes[] = {
    "Unknown",
    "Mandatory upgrade",
    "Advisory upgrade",
    "System bulletin",
    "Top o' the world!"};
  static int codeslen = 5;

  char *amsg;
  unsigned short id;
  va_list ap;
  char msg[1024];
  
  va_start(ap, command);
  id = va_arg(ap, u_int);
  msg = va_arg(ap, char *);
  va_end(ap);

  printf("faimtest: motd: %s (%d / %s)\n", msg, id, 
	 (id < codeslen)?codes[id]:"unknown");
*/
  //aim_0001_0020(sess, command->conn);

  return 1;
}

int at_parse_connerr(struct aim_session_t *sess, 
                     struct command_rx_struct *command, ...)
{
  va_list ap;
  unsigned short code;
  char *msg = NULL;
  at_session s;

  s = (at_session)sess->aux_data;
  va_start(ap, command);
  code = va_arg(ap, u_int);
  msg = va_arg(ap, char *);
  va_end(ap);

  /* XXX End the session */
  aim_conn_kill(sess, &command->conn); /* this will break the main loop */

  return 1;
}

int at_flapversion(struct aim_session_t *sess, 
                   struct command_rx_struct *command, ...)
{

  return 1;
}

int at_conncomplete(struct aim_session_t *sess, 
                    struct command_rx_struct *command, ...)
{
  va_list ap;
  struct aim_conn_t *conn;

  va_start(ap, command);
  conn = va_arg(ap, struct aim_conn_t *);
  va_end(ap);
  
  return 1;
}

#if 0
at_session at_session_find_aim(int fd)
{
    struct aim_conn_t *cur;

    for (cur = sess->connlist; cur; cur = cur->next) {
        if (cur->status & AIM_CONN_STATUS_INPROGRESS) {
            FD_SET(cur->fd, &wfds);
            haveconnecting++;
        }
        FD_SET(cur->fd, &fds);
        if (cur->fd > maxfd)
            maxfd = cur->fd;
    }
}
#endif

#if 0
ssize_t at_session_handle_read(mio m, void* buf, size_t count)
{
    at_mio am;
    at_session s;
    struct aim_conn_t *conn;
    int flags;

    am = (at_mio)m->cb_arg;

    flags = fcntl(m->fd, F_GETFL, 0);
    if(flags & O_NONBLOCK)
    {
        flags &= ~O_NONBLOCK;
        if(fcntl(m->fd, F_SETFL, flags)<0)
            log_debug(ZONE, "[AT] Error calling F_SETFL to block");
    }


    /*log_debug(ZONE, "[AT] Session MIO handle read");*/
    errno = 0;
    if (aim_get_command(am->ass, am->conn) >= 0)
    {
        aim_rxdispatch(am->ass);
        return 1;
    } else {
        log_debug(ZONE, "[AT] Error handling read!");
        aim_conn_kill(am->ass, &am->conn);
        mio_close(m);
    //    at_session_end(s);
        return -1;
    }
}

ssize_t at_session_handle_write(mio m, const void* buf, size_t count)
{
    /*log_debug(ZONE, "[AT] Session MIO handle write");*/
    errno = 0;
    return count;
}

void at_handle_mio(mio m, int flags, void *arg, char *buffer, int bufsz)
{
    struct aim_conn_t *cur;
    at_mio am;
    at_session s;
    int sflags;

    am = (at_mio)arg;

    if(flags == MIO_NEW)
        return;

    if(flags == MIO_CLOSED || flags == MIO_ERROR)
    {
        log_debug(ZONE, "[AT] socket %d closed", m->fd);
        pool_free(am->p);
        if(am->s->loggedin == 1)
            at_session_end(am->s);
        return;
    }
}
#endif

void at_debugcb(struct aim_session_t *sess, int level, const char *format, va_list va)
{
    vfprintf(stderr, format, va);
    
    return;
}

at_session at_session_create(ati ti, xmlnode aim_data, jpacket jp)
{
    char *user, *pword;
    jid j, from;
    pool p;
    mio m = NULL;
    at_session new;
    struct aim_session_t *ass;
    struct aim_conn_t *authconn = NULL;
    jid jput;
    pth_attr_t attr;
    xmlnode fpres;

    user = xmlnode_get_attrib(aim_data, "id");
    pword = xmlnode_get_attrib(aim_data, "pass");

    j = jp->from;
    from = jp->to;

    new = at_session_find_by_jid(ti, j);
    if(new!=NULL)
    {
        log_debug(ZONE, "[AIM] The goober is already online");
        return NULL;
    }
    log_debug(ZONE, "[AT] Going to start session for %s", jid_full(j));
    ass = malloc(sizeof(struct aim_session_t));
	//aim_session_init(ass, 0);
	aim_session_init(ass, AIM_SESS_FLAGS_NONBLOCKCONNECT, 0);
    aim_setdebuggingcb(ass, at_debugcb);

    /* Set us to immediate queue mode */
    aim_tx_setenqueue(ass, AIM_TX_IMMEDIATE, NULL);

    authconn = aim_newconn(ass, AIM_CONN_TYPE_AUTH, FAIM_LOGIN_SERVER);
    if(authconn == NULL)
    {
        xmlnode err;
        
        /* Let the user know */
        err = xmlnode_new_tag("message");
        xmlnode_put_attrib(err, "to", jid_full(j));
        xmlnode_put_attrib(err, "type", "error");
        xmlnode_put_attrib(err, "from", jid_full(from));
        jutil_error(err, (terror){500, "Error connecting to the AIM authentication server"});
        deliver(dpacket_new(err), ti->i);

        /* Break it down */
        log_error("[AIM]", "Error connecting to aims authentication server.\n");
        return NULL;
    } 
    else if (authconn->fd == -1)
    {
        xmlnode err;
        
        /* Let the user know */
        err = xmlnode_new_tag("message");
        xmlnode_put_attrib(err, "to", jid_full(j));
        xmlnode_put_attrib(err, "type", "error");
        xmlnode_put_attrib(err, "from", jid_full(from));
        
        if (authconn->status & AIM_CONN_STATUS_RESOLVERR)
        {
            jutil_error(err, (terror){500, "Could not resolve AIM authorizer"});
            log_error("[AIM]", "at: could not resolve authorizer name");
        }
        else if (authconn->status & AIM_CONN_STATUS_CONNERR)
        {
            jutil_error(err, (terror){500, "Could not connect to AIM authorizer"});
            log_error("[AIM]", "at: could not connect to authorizer");
        }
        deliver(dpacket_new(err), ti->i);

        /* Clean up and get out */
        aim_conn_kill(ass, &authconn);
        return NULL;
    }
  	else
    {


        aim_conn_addhandler(ass, authconn, AIM_CB_FAM_SPECIAL, 
                            AIM_CB_SPECIAL_FLAPVER, at_flapversion, 0);
        aim_conn_addhandler(ass, authconn, AIM_CB_FAM_SPECIAL, 
                            AIM_CB_SPECIAL_CONNCOMPLETE, at_conncomplete, 0);
        aim_conn_addhandler(ass, authconn, 0x0017, 0x0007, 
                            at_parse_login, 0);
        aim_conn_addhandler(ass, authconn, 0x0017, 0x0003, 
                            at_parse_authresp, 0);
    
        //m = mio_new(authconn->fd, &at_handle_mio, authconn, mio_handlers_new(at_session_handle_read, at_session_handle_write, NULL));
  
        //m->rated = 0;
        p = pool_new();
        new = pmalloc(p, sizeof(_at_session));
        new->m = NULL;
        new->p = p;
        new->ti = ti;
    	new->mp_to = pth_msgport_create("aimsession_to");
        new->cur=jid_new(new->p, jid_full(j));
        new->from=jid_new(new->p, jid_full(from));
        new->ass = ass;
        new->exit_flag = 0;
        new->loggedin = 0;
        new->status = NULL;
        new->away = 0;
        new->buddies = xhash_new(137);
        new->profile = NULL;
        new->at_PPDB = NULL;
        new->screenname = pstrdup(new->p, user);
        new->password = pstrdup(new->p, pword);
        new->ass->aux_data = (void *)new;

        fpres = jutil_presnew(JPACKET__AVAILABLE, ti->i->id, "Online");
        xmlnode_put_attrib(fpres, "from", jid_full(new->cur));
        new->at_PPDB = ppdb_insert(new->at_PPDB, new->cur, fpres);
        xmlnode_free(fpres);
        /*
        am = pmalloco(new->p, sizeof(_at_mio));
        am->p = new->p;
        am->s = new;
        am->conn = authconn;
        am->ass = ass;
        */

        /* mio_reset(m, &at_handle_mio, am); */

        /* Ok everything is good, put this in the session hash */
        jput = jid_new(new->p, jid_full(new->cur));
        jid_set(jput, NULL, JID_RESOURCE);
        xhash_put(ti->session__list, jid_full(jput), new);
    
        log_debug(ZONE, "[AT] User Login: %s", jid_full(new->cur));

        aim_request_login(ass, authconn, user);
        
        attr = pth_attr_new();
        new->tid = pth_spawn(attr, at_session_main, (void *)new);
    	return new;
    }

    return NULL;
}

void *at_session_main(void *arg)
{
    at_session s;
    ati ti;
	pth_event_t to;
	jpq q;
	xmlnode x;
	struct aim_conn_t *waitingconn = NULL;
    int status;

    s = (at_session)arg;
    ti = s->ti;
	to = pth_event(PTH_EVENT_MSG, s->mp_to);

	/* Do stuff */
	log_debug(ZONE, "[AIM] In our main session");
	while ((waitingconn = _aim_select(s->ass, to, &status)) >= 0)
    {
		if(s->exit_flag)
			break;

        switch(status)
        {
            case -1:
                /* There was an error */
                s->exit_flag = 1;
                break;
            case 0:
                /* Nothing to do */
                break;
            case 1:
			    log_debug(ZONE, "[AIM] Flushing outgoing queue");
    			aim_tx_flushqueue(s->ass);
                break;
            case 2:
    		    if (aim_get_command(s->ass, waitingconn) < 0)
                {
	    			/* XXX do something like send an error */
		    		log_debug(ZONE,"[AIM] connection error (type 0x%04x:0x%04x)", waitingconn->type, waitingconn->subtype);
                    aim_conn_kill(s->ass, &waitingconn);
                    if (!aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS))
	                      log_debug(ZONE,"major connection error");
			    	s->exit_flag = 1;
    			}
	        	else
		    	{
			    	aim_rxdispatch(s->ass);
    			}
                break;
            case 3:
		        while(1)
        		{
		        	q = (jpq)pth_msgport_get(s->mp_to);
        			if(q == NULL)
                    {
		        		break;
                    }
                    at_aim_session_parser(s, q->p);
                    if(s->exit_flag)
                        break;
        		} 
                break;
        }

		if(s->exit_flag)
			break;
    }

	/* Close up */
	log_debug(ZONE,"[AIM] Closing up a main thread");

	pth_event_free(to, PTH_FREE_ALL);

	at_session_end(s);

    pth_exit(NULL);

	return NULL;
}
	
void at_aim_session_parser(at_session s, jpacket jp)
{
    ati ti;
    int freed;
    char *ns;

    ti = s->ti;
    log_debug(ZONE, "[AIM] Parsing Packet on sessions");
        
    if(s->exit_flag > 0)
    {
        xmlnode_free(jp->x);
        return;
    }
    switch(jp->type)
    {
        case JPACKET_IQ:
            if(NSCHECK(xmlnode_get_tag(jp->x, "query"),NS_REGISTER))
            {
                /* Register the user */
                freed = at_register(ti, jp);
                break;
            }
            ns = xmlnode_get_attrib(xmlnode_get_tag(jp->x, "query"), "xmlns");
            freed = at_run_iqcb(ti, ns, jp);
            if(freed < 0)
            {
                /* XXX We could dup here and then return 0... */
                jutil_error(jp->x, TERROR_NOTFOUND);
                deliver(dpacket_new(jp->x), ti->i);
                freed = 1;
            }
            break;
	    case JPACKET_MESSAGE:
            /* Put the message on the line for the session to handle */
            at_session_deliver(s, jp->x, jp->to);
            freed = 1;
            break;
        case JPACKET_S10N:
            /* Handle a S10N packet based on their subtype */
            log_debug(ZONE, "[AT] We got a s10n packet");
            freed = at_session_s10n(s, jp);
            break;
		case JPACKET_PRESENCE:
            freed = at_session_pres(s, jp);
            break;
        default:
            /* XXX do more!? */
            /*jutil_error(jp->x, TERROR_NOTACCEPTABLE);
            deliver(dpacket_new(jp->x), ti->i);
            */
            xmlnode_free(jp->x);
            freed = 1;
            break;
	}

    if(freed==0)
        xmlnode_free(jp->x);
}

void at_session_deliver(at_session s, xmlnode x, jid from)
{
    /* XXX We need to check for the XHTML version of the msg */
    char *body;

    body = UTF8_to_str(xmlnode_pool(x),xmlnode_get_tag_data(x, "body"));

    if(body == NULL || from->user == NULL)
        return;

    log_debug(ZONE, "[AIM] Sending a Message");
    aim_send_im(s->ass, aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS), 
                from->user, 0, body, strlen(body));
    xmlnode_free(x);
    return;
}

void _at_session_save_buddies(xht ht, const char *key, void *data, void *arg)
{
    xmlnode item;
    xmlnode roster;
    char *user;

    roster = (xmlnode)arg;
    user = (char *)key;

    item = xmlnode_insert_tag(roster, "item");
    xmlnode_put_attrib(item, "name", user);

    return;
}

void at_session_end(at_session s)
{
    ati ti;
    xmlnode x;
    xmlnode xset;
    int ret;
    jid jend;

    if(s == NULL)
        log_warn(ZONE,"NULL Session told to end!");
    else
    {
        ti = s->ti;

        log_debug(ZONE, "[AT] Session (%s) told to end.", jid_full(s->cur));
        jend = jid_new(s->p, jid_full(s->cur));
        jid_set(jend, NULL, JID_RESOURCE);
        xhash_zap(ti->session__list, jid_full(jend));
        x = jutil_presnew(JPACKET__UNAVAILABLE, jid_full(s->cur), "Disconnected");
        xmlnode_put_attrib(x, "from", jid_full(s->from));
        deliver(dpacket_new(x), ti->i);

        xset = xmlnode_new_tag("buddies");
        xhash_walk(s->buddies, &_at_session_save_buddies, xset);
        log_debug(ZONE, "Saving buddies: %s", xmlnode2str(xset));
        ret = at_xdb_set(ti, ti->i->id, s->cur, xset, AT_NS_ROSTER);
        if(ret == 1)
            xmlnode_free(xset);

        pth_msgport_destroy(s->mp_to);
        aim_logoff(s->ass);
        aim_session_kill(s->ass);
        log_debug(ZONE,"[AT] unlinking session");
        at_session_unlink_buddies(s);
        xhash_free(s->buddies);
        log_warn(ZONE, "[AT] Closing down session for %s", jid_full(s->cur));
        ppdb_free(s->at_PPDB);
        free(s->ass);
        pool_free(s->p);
    }
    return;
}

void _at_session_unlink_buddies(xht ht, const char *key, void *data, void *arg)
{
    ati ti;
    at_session s;
    xmlnode p;
    at_buddy buddy;

    s = (at_session)arg;
    buddy = (at_buddy)data;
    ti = s->ti;

    p = jutil_presnew(JPACKET__UNAVAILABLE, jid_full(s->cur), "Transport Disconnected");
    xmlnode_put_attrib(p, "from", jid_full(buddy->full));

    deliver(dpacket_new(p), ti->i);

    return;
}
void at_session_unlink_buddies(at_session s)
{
    at_buddy cur, next;

    xhash_walk(s->buddies, &_at_session_unlink_buddies, s);
}

int session_walk(void *user_data, const void *key, void *data)
{
    log_debug(ZONE, "[AT] Walking key: %s data: %x", key, data);
    return 0;
}

at_session at_session_find_by_jid(ati ti, jid j)
{
    at_session s;
    char *res;

    if(j == NULL)
        return NULL;

    res = j->resource;
    jid_set(j, NULL, JID_RESOURCE);
    log_debug(ZONE, "[AT] Finding session for %s", jid_full(j));
    s = (at_session)xhash_get(ti->session__list, jid_full(j));
    j->resource = res;
    /* XXX LEAK ALERT! This is a jid handling problem that we're working on */
    j->full = NULL;

    return s;
}

struct aim_conn_t *_aim_select(struct aim_session_t *sess, 
                               pth_event_t ev, 
                               int *status)
{
    struct aim_conn_t *cur;
    fd_set fds, wfds;
    int maxfd = 0;
    int i, haveconnecting = 0;

    faim_mutex_lock(&sess->connlistlock);
    if (sess->connlist == NULL) {
        faim_mutex_unlock(&sess->connlistlock);
        *status = -1;
        return NULL;
    }
    faim_mutex_unlock(&sess->connlistlock);


    FD_ZERO(&fds);
    FD_ZERO(&wfds);
    maxfd = 0;

    faim_mutex_lock(&sess->connlistlock);
    for (cur = sess->connlist; cur; cur = cur->next) {
        if (cur->status & AIM_CONN_STATUS_INPROGRESS) {
            FD_SET(cur->fd, &wfds);
            haveconnecting++;
        }
        FD_SET(cur->fd, &fds);
        if (cur->fd > maxfd)
            maxfd = cur->fd;
    }
    faim_mutex_unlock(&sess->connlistlock);
    
    /* 
    * If we have data waiting to be sent, return immediatly
    */
    if (!haveconnecting && (sess->queue_outgoing != NULL)) {
        *status = 1;
        return NULL;
    } 

    if ((i = pth_select_ev(maxfd+1, &fds, &wfds, NULL, NULL, ev))>=1) {
        faim_mutex_lock(&sess->connlistlock);
        for (cur = sess->connlist; cur; cur = cur->next) {
            if (FD_ISSET(cur->fd, &fds) || 
                ((cur->status & AIM_CONN_STATUS_INPROGRESS) && 
                FD_ISSET(cur->fd, &wfds))) 
            {
	            *status = 2;
	            faim_mutex_unlock(&sess->connlistlock);
	            return cur;
            }
        }
        *status = 0; /* shouldn't happen */
    } else if ((i == -1) && (errno == EINTR))/* treat inerpts as a timeout */
        *status = 0;
    else
        *status = i; /* can be 0 or -1 */

    if(pth_event_occurred(ev))
    {
        /* Possible msgport event (should be, no timeout) */
        *status = 3;
        return (struct aim_conn_t *)1;
    }

    faim_mutex_unlock(&sess->connlistlock);
    return NULL;  /* no waiting or error, return */
}
